from user_streams import BACKEND_SETTING_NAME
from user_streams.tests import StreamStorageTestMixin
from user_streams.utils import TestCase, override_settings


BACKEND_SETTINGS = {BACKEND_SETTING_NAME: 'user_streams.backends.many_to_many.ManyToManyDatabaseBackend'}


@override_settings(**BACKEND_SETTINGS)
class ManyToManyDatabaseBackendTestCase(TestCase, StreamStorageTestMixin):
    pass



from .models import StreamItem


class ManyToManyDatabaseBackend(object):

    def add_stream_item(self, users, content, created_at):
        item = StreamItem.objects.create(content=content, created_at=created_at)
        item.users.add(*[user.pk for user in users])

    def get_stream_items(self, user):
        return StreamItem.objects.filter(users=user)




class DummyStreamItem(object):

    def __init__(self, content, created_at):
        self.content = content
        self.created_at = created_at


class MemoryStorage(object):

    def __init__(self):
        self.streams = {}

    def add_stream_item(self, users, content, created_at):
        stream_item = DummyStreamItem(content, created_at)
        for user in users:
            if user in self.streams:
                self.streams[user].insert(0, stream_item)
            else:
                self.streams[user] = [stream_item]

    def get_stream_items(self, user):
        return self.streams.get(user, [])

    def flush(self):
        self.streams = {}


storage = MemoryStorage()


class DummyBackend(object):

    """
    A dummy storage backend that stores user streams in memory.
    Only used for testing purposes.
    """

    def add_stream_item(self, users, content, created_at):
        storage.add_stream_item(users, content, created_at)

    def get_stream_items(self, user):
        return storage.get_stream_items(user)

    def flush(self):
        storage.flush()




DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    },
}

INSTALLED_APPS = [
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'user_streams',
    'user_streams.backends.single_table',
    'user_streams.backends.many_to_many',
    'user_streams.backends.redis',
]



try:
    from redis import Redis
except ImportError:
    raise ImportError('Please install the redis-py module (pip install redis)')


from datetime import datetime
from django.utils.encoding import smart_str, smart_unicode
import time
from uuid import uuid4


KEY_PREFIX_SETTING_NAME = 'USER_STREAMS_REDIS_KEY_PREFIX'
DEFAULT_KEY_PREFIX = 'user_streams'
CLIENT_ARGUMENTS_SETTING_NAME = 'USER_STREAMS_REDIS_CLIENT_ARGUMENTS'


def get_redis_client():
    from django.conf import settings
    client_arguments = getattr(settings, CLIENT_ARGUMENTS_SETTING_NAME, {})
    return Redis(**client_arguments)


def create_key(key):
    from django.conf import settings
    prefix = getattr(settings, KEY_PREFIX_SETTING_NAME, DEFAULT_KEY_PREFIX)
    return "%s:%s" % (prefix, key)


def add_header(content):
    """
    We need to add a unique header to each message, as duplicate items
    will otherwise be overwritten
    """
    return uuid4().hex + smart_str(content)


def remove_header(content):
    return smart_unicode(content[32:])


class RedisBackend(object):

    def __init__(self):
        self.redis_client = get_redis_client()

    def add_stream_item(self, users, content, created_at):
        content = add_header(content)
        for user in users:
            key = create_key('user:%s' % user.pk)
            timestamp = time.mktime(created_at.timetuple())
            self.redis_client.zadd(key, content, timestamp)

    def get_stream_items(self, user):
        return LazyResultSet(user)


class LazyResultSet(object):

    def __init__(self, user):
        self.key = create_key('user:%s' % user.pk)
        self.redis_client = get_redis_client()

    def __len__(self):
        return self.redis_client.zcard(self.key)

    def __getitem__(self, item):
        if isinstance(item, slice):
            start = item.start or 0
            stop = item.stop or -1
        else:
            start = item
            stop = item

        result = self.redis_client.zrange(self.key, start, stop, desc=True, withscores=True)
        for content, timestamp in result:
            created_at = datetime.fromtimestamp(timestamp)
            content = remove_header(content)
            return StreamItem(content, created_at)


class StreamItem(object):

    def __init__(self, content, created_at):
        self.content = content
        self.created_at = created_at



# -*- coding: utf-8 -*-


from datetime import datetime, timedelta
from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured


from user_streams import BACKEND_SETTING_NAME, get_backend, add_stream_item, get_stream_items
from user_streams.backends.dummy import DummyBackend
from user_streams.utils import TestCase, override_settings


DUMMY_BACKEND_SETTINGS = {BACKEND_SETTING_NAME: 'user_streams.backends.dummy.DummyBackend'}


class GetBackendTestCase(TestCase):

    def test_missing_setting(self):
        with self.assertRaises(ImproperlyConfigured):
            get_backend()

    def test_invalid_backend_path(self):
        settings = {BACKEND_SETTING_NAME: 'invalid'}
        with self.settings(**settings):
            with self.assertRaises(ImproperlyConfigured):
                get_backend()

    def test_incorrect_backend_path(self):
        settings = {BACKEND_SETTING_NAME: 'foo.bar.invalid.InvalidClass'}
        with self.settings(**settings):
            with self.assertRaises(ImproperlyConfigured):
                get_backend()

    def test_correct_backend_returned(self):
        with self.settings(**DUMMY_BACKEND_SETTINGS):
            backend = get_backend()
            self.assertTrue(isinstance(backend, DummyBackend))


class StreamStorageTestMixin(object):

    """
    A mixin providing a set of test cases that can be run to test
    any backend. Note that the backend MUST be emptied (all messages
    should be removed) between each test. If a database backend
    is being tested, this will happen automatically. Otherwise, you
    are responsible for deleting all the messages between tests.
    """

    def test_single_user(self):
        user = User.objects.create()
        content = 'Test message'

        add_stream_item(user, content)

        items = get_stream_items(user)
        self.assertEqual(len(items), 1)
        item = items[0]
        self.assertEqual(item.content, content)

    def test_multiple_users(self):
        user_1 = User.objects.create(username='test1')
        user_2 = User.objects.create(username='test2')
        user_3 = User.objects.create(username='test3')
        content = 'Broadcast message'

        add_stream_item(User.objects.all(), content)

        for user in user_1, user_2, user_3:
            self.assertEqual(get_stream_items(user)[0].content, content)

    def test_message_ordering(self):
        user = User.objects.create()
        now = datetime.now()

        add_stream_item(user, 'Message 1', created_at=now)
        add_stream_item(user, 'Message 2', created_at=now + timedelta(minutes=1))
        add_stream_item(user, 'Message 3', created_at=now + timedelta(minutes=2))

        stream_items = get_stream_items(user)

        self.assertEqual(stream_items[0].content, 'Message 3')
        self.assertEqual(stream_items[1].content, 'Message 2')
        self.assertEqual(stream_items[2].content, 'Message 1')

    def test_identical_messages(self):
        """Check that identical messages are handled properly. Mostly
        an issue for the Redis backend (which uses sets to store messages)"""
        user = User.objects.create()
        message = 'Test message'

        add_stream_item(user, message)
        add_stream_item(user, message)

        items = get_stream_items(user)
        self.assertEqual(len(items), 2)

    def test_unicode_handled_properly(self):
        user = User.objects.create()
        message = u'☃'

        add_stream_item(user, message)

        items = get_stream_items(user)
        self.assertEqual(items[0].content, message)



@override_settings(**DUMMY_BACKEND_SETTINGS)
class DummyBackendStreamTestCase(TestCase, StreamStorageTestMixin):

    def setUp(self):
        dummy_backend = get_backend()
        dummy_backend.flush()



from datetime import datetime
from django.core.exceptions import ImproperlyConfigured
from django.utils.importlib import import_module


BACKEND_SETTING_NAME = 'USER_STREAMS_BACKEND'
USE_UTC_SETTING_NAME = 'USER_STREAMS_USE_UTC'


def get_backend():
    """
    Return the stream storage backend configured in the settings
    """
    from django.conf import settings
    backend_path = getattr(settings, BACKEND_SETTING_NAME, None)
    if not backend_path:
        raise ImproperlyConfigured('No user stream storage backend has been configured. Please set %s correctly' % BACKEND_SETTING_NAME)

    try:
        module_name, class_name = backend_path.rsplit('.', 1)
    except ValueError:
        raise ImproperlyConfigured('%s is not a valid value for the %s setting' % (backend_path, BACKEND_SETTING_NAME))

    try:
        module = import_module(module_name)
    except ImportError, e:
        raise ImproperlyConfigured('Error importing user stream backend %s: %s' % (backend_path, e))

    try:
        cls = getattr(module, class_name)
    except AttributeError:
        raise ImproperlyConfigured('Error importing user stream backend class %s' % backend_path)

    return cls()


def create_iterable(item_or_iterable):
    """
    If the argument is iterable, just return it. Otherwise, return a list
    containing the item.
    """
    try:
        iter(item_or_iterable)
        return item_or_iterable
    except TypeError:
        return [item_or_iterable]


def now():
    """
    Return a datetime object representing the current time
    """
    from django.conf import settings
    use_utc = getattr(settings, USE_UTC_SETTING_NAME, False)
    if use_utc:
        return datetime.utcnow()
    else:
        return datetime.now()


def add_stream_item(user_or_users, content, created_at=None):
    """
    Add a single message to the stream of one or more users.
    """
    backend = get_backend()
    users = create_iterable(user_or_users)
    created_at = created_at or now()
    backend.add_stream_item(users, content, created_at)


def get_stream_items(user):
    """
    Retrieve the stream for a single user.
    """
    backend = get_backend()
    return backend.get_stream_items(user)



from user_streams import BACKEND_SETTING_NAME
from user_streams.tests import StreamStorageTestMixin
from user_streams.utils import TestCase, override_settings

from . import Redis, KEY_PREFIX_SETTING_NAME


KEY_PREFIX = 'redis_backend_tests'
BACKEND_SETTINGS = {
    BACKEND_SETTING_NAME: 'user_streams.backends.redis.RedisBackend',
    KEY_PREFIX_SETTING_NAME: KEY_PREFIX,
}


@override_settings(**BACKEND_SETTINGS)
class RedisBackendTestCase(TestCase, StreamStorageTestMixin):

    def tearDown(self):
        client = Redis()
        keys = client.keys('%s*' % KEY_PREFIX)
        if keys:
            client.delete(*keys)



from .models import StreamItem


class SingleTableDatabaseBackend(object):

    def add_stream_item(self, users, content, created_at):
        for user in users:
            StreamItem.objects.create(user=user, content=content, created_at=created_at)

    def get_stream_items(self, user):
        return StreamItem.objects.filter(user=user)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import subprocess
from setuptools import setup, find_packages


version = '0.2.0'


if sys.argv[-1] == 'publish':
    subprocess.call(['python', 'setup.py', 'sdist', 'upload'])
    print "You probably want to also tag the version now:"
    print "  git tag -a %s -m 'Tag version %s'" % (version, version)
    print "  git push --tags"
    sys.exit()


setup(
    name='django-user-streams',
    version=version,
    description='Simple, fast user news feeds for Django',
    author='Jamie Matthews',
    url='https://github.com/dabapps/django-user-streams',
    packages=find_packages(exclude=['testproject']),
    license='BSD',
)



from django.db import models


class StreamItem(models.Model):

    user = models.ForeignKey('auth.User', related_name='+')
    content = models.TextField()
    created_at = models.DateTimeField()

    class Meta:
        ordering = ['-created_at']
